"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseDeclarations = void 0;
const names_js_1 = require("./names.js");
const expression_js_1 = require("./expression.js");
const util_js_1 = require("./util.js");
const values_js_1 = require("./values.js");
function parseDeclarations(ctx) {
    const { source } = ctx;
    let pos = (0, util_js_1.whitespaces)(source, 0);
    const declarations = [];
    loop: while (source[pos] === '.') {
        const keyword = (0, names_js_1.parseNameValue)(source, pos + 1);
        let decl;
        switch (keyword) {
            case '':
            case 'match':
                break loop;
            case 'input':
                decl = parseInputDeclaration(ctx, pos);
                break;
            case 'local':
                decl = parseLocalDeclaration(ctx, pos);
                break;
            default:
                decl = parseReservedStatement(ctx, pos, '.' + keyword);
        }
        declarations.push(decl);
        pos = decl.end;
        pos += (0, util_js_1.whitespaces)(source, pos);
    }
    return { declarations, end: pos };
}
exports.parseDeclarations = parseDeclarations;
function parseInputDeclaration(ctx, start) {
    //
    let pos = start + 6; // '.input'
    const keyword = { start, end: pos, value: '.input' };
    pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    const value = parseDeclarationValue(ctx, pos);
    if (value.type === 'expression') {
        if (value.markup || value.arg?.type !== 'variable') {
            ctx.onError('bad-input-expression', value.start, value.end);
        }
    }
    return { type: 'input', start, end: value.end, keyword, value };
}
function parseLocalDeclaration(ctx, start) {
    const { source } = ctx;
    let pos = start + 6; // '.local'
    const keyword = { start, end: pos, value: '.local' };
    const ws = (0, util_js_1.whitespaces)(source, pos);
    pos += ws;
    if (ws === 0)
        ctx.onError('missing-syntax', pos, ' ');
    let target;
    if (source[pos] === '$') {
        target = (0, values_js_1.parseVariable)(ctx, pos);
        pos = target.end;
    }
    else {
        const junkStart = pos;
        const junkEndOffset = source.substring(pos).search(/[\t\n\r ={}]/);
        pos = junkEndOffset === -1 ? source.length : pos + junkEndOffset;
        target = {
            type: 'junk',
            start: junkStart,
            end: pos,
            source: source.substring(junkStart, pos)
        };
        ctx.onError('missing-syntax', junkStart, '$');
    }
    pos += (0, util_js_1.whitespaces)(source, pos);
    let equals;
    if (source[pos] === '=') {
        equals = { start: pos, end: pos + 1, value: '=' };
        pos += 1;
    }
    else {
        ctx.onError('missing-syntax', pos, '=');
    }
    pos += (0, util_js_1.whitespaces)(source, pos);
    const value = parseDeclarationValue(ctx, pos);
    return {
        type: 'local',
        start,
        end: value.end,
        keyword,
        target,
        equals,
        value
    };
}
function parseReservedStatement(ctx, start, keyword) {
    let pos = start + keyword.length;
    pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    const body = (0, expression_js_1.parseReservedBody)(ctx, pos);
    let end = body.end;
    pos = end + (0, util_js_1.whitespaces)(ctx.source, end);
    const values = [];
    while (ctx.source[pos] === '{') {
        if (ctx.source.startsWith('{{', pos))
            break;
        const value = (0, expression_js_1.parseExpression)(ctx, pos);
        values.push(value);
        end = value.end;
        pos = end + (0, util_js_1.whitespaces)(ctx.source, end);
    }
    if (values.length === 0)
        ctx.onError('missing-syntax', end, '{');
    return {
        type: 'reserved-statement',
        start,
        end,
        keyword: { start, end: keyword.length, value: keyword },
        body,
        values
    };
}
function parseDeclarationValue(ctx, start) {
    const { source } = ctx;
    if (source[start] === '{')
        return (0, expression_js_1.parseExpression)(ctx, start);
    const junkEndOffset = source.substring(start).search(/\.[a-z]|{{/);
    const end = junkEndOffset === -1 ? source.length : start + junkEndOffset;
    ctx.onError('missing-syntax', start, '{');
    return { type: 'junk', start, end, source: source.substring(start, end) };
}
