"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseCST = exports.ParseContext = void 0;
const errors_js_1 = require("../errors.js");
const declarations_js_1 = require("./declarations.js");
const expression_js_1 = require("./expression.js");
const util_js_1 = require("./util.js");
const values_js_1 = require("./values.js");
class ParseContext {
    errors = [];
    resource;
    source;
    constructor(source, opt) {
        this.resource = opt?.resource ?? false;
        this.source = source;
    }
    onError(type, start, end) {
        let err;
        if (type === 'missing-syntax') {
            const exp = String(end);
            err = new errors_js_1.MessageSyntaxError(type, start, start + exp.length, exp);
        }
        else {
            err = new errors_js_1.MessageSyntaxError(type, start, Number(end));
        }
        this.errors.push(err);
    }
}
exports.ParseContext = ParseContext;
/**
 * Parse the string syntax representation of a message into
 * its corresponding CST representation.
 *
 * @beta
 */
function parseCST(source, opt) {
    const ctx = new ParseContext(source, opt);
    if (source.startsWith('.')) {
        const { declarations, end: pos } = (0, declarations_js_1.parseDeclarations)(ctx);
        return source.startsWith('.match', pos)
            ? parseSelectMessage(ctx, pos, declarations)
            : parsePatternMessage(ctx, pos, declarations, true);
    }
    else {
        return parsePatternMessage(ctx, 0, [], source.startsWith('{{'));
    }
}
exports.parseCST = parseCST;
function parsePatternMessage(ctx, start, declarations, complex) {
    const pattern = parsePattern(ctx, start, complex);
    let pos = pattern.end;
    pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    if (pos < ctx.source.length) {
        ctx.onError('extra-content', pos, ctx.source.length);
    }
    return complex
        ? { type: 'complex', declarations, pattern, errors: ctx.errors }
        : { type: 'simple', pattern, errors: ctx.errors };
}
function parseSelectMessage(ctx, start, declarations) {
    let pos = start + 6; // '.match'
    const match = { start, end: pos, value: '.match' };
    pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    const selectors = [];
    while (ctx.source[pos] === '{') {
        const sel = (0, expression_js_1.parseExpression)(ctx, pos);
        const body = sel.markup ?? sel.annotation;
        if (body && body.type !== 'function') {
            ctx.onError('bad-selector', body.start, body.end);
        }
        selectors.push(sel);
        pos = sel.end;
        pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    }
    if (selectors.length === 0) {
        ctx.onError('empty-token', pos, pos + 1);
    }
    const variants = [];
    pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    while (pos < ctx.source.length) {
        const variant = parseVariant(ctx, pos);
        variants.push(variant);
        pos = variant.end;
        pos += (0, util_js_1.whitespaces)(ctx.source, pos);
    }
    if (pos < ctx.source.length) {
        ctx.onError('extra-content', pos, ctx.source.length);
    }
    return {
        type: 'select',
        declarations,
        match,
        selectors,
        variants,
        errors: ctx.errors
    };
}
function parseVariant(ctx, start) {
    let pos = start;
    const keys = [];
    while (pos < ctx.source.length) {
        const ws = (0, util_js_1.whitespaces)(ctx.source, pos);
        pos += ws;
        const ch = ctx.source[pos];
        if (ch === '{')
            break;
        if (pos > start && ws === 0)
            ctx.onError('missing-syntax', pos, "' '");
        const key = ch === '*'
            ? ({ type: '*', start: pos, end: pos + 1 })
            : (0, values_js_1.parseLiteral)(ctx, pos, true);
        if (key.end === pos)
            break; // error; reported in pattern.errors
        keys.push(key);
        pos = key.end;
    }
    const value = parsePattern(ctx, pos, true);
    return { start, end: value.end, keys, value };
}
function parsePattern(ctx, start, quoted) {
    let pos = start;
    if (quoted) {
        if (ctx.source.startsWith('{{', pos)) {
            pos += 2;
        }
        else {
            ctx.onError('missing-syntax', start, '{{');
            return { start, end: start, body: [] };
        }
    }
    const body = [];
    loop: while (pos < ctx.source.length) {
        switch (ctx.source[pos]) {
            case '{': {
                const ph = (0, expression_js_1.parseExpression)(ctx, pos);
                body.push(ph);
                pos = ph.end;
                break;
            }
            case '}':
                break loop;
            default: {
                const tx = (0, values_js_1.parseText)(ctx, pos);
                body.push(tx);
                pos = tx.end;
            }
        }
    }
    if (quoted) {
        const q0 = { start, end: start + 2, value: '{{' };
        let braces;
        if (ctx.source.startsWith('}}', pos)) {
            braces = [q0, { start: pos, end: pos + 2, value: '}}' }];
            pos += 2;
        }
        else {
            braces = [q0];
            ctx.onError('missing-syntax', pos, '}}');
        }
        return { braces, start, end: pos, body };
    }
    return { start, end: pos, body };
}
