"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseVariable = exports.parseQuotedLiteral = exports.parseLiteral = exports.parseText = void 0;
const names_js_1 = require("./names.js");
// Text ::= (TextChar | TextEscape)+
// TextChar ::= AnyChar - ('{' | '}' | Esc)
// AnyChar ::= [#x0-#x10FFFF]
// Esc ::= '\'
// TextEscape ::= Esc Esc | Esc '{' | Esc '}'
function parseText(ctx, start) {
    let value = '';
    let pos = start;
    let i = start;
    loop: for (; i < ctx.source.length; ++i) {
        switch (ctx.source[i]) {
            case '\\': {
                const esc = parseEscape(ctx, 'text', i);
                if (esc) {
                    value += ctx.source.substring(pos, i) + esc.value;
                    i += esc.length;
                    pos = i + 1;
                }
                break;
            }
            case '{':
            case '}':
                break loop;
            case '\n':
                if (ctx.resource) {
                    const nl = i;
                    let next = ctx.source[i + 1];
                    while (next === ' ' || next === '\t') {
                        i += 1;
                        next = ctx.source[i + 1];
                    }
                    if (i > nl) {
                        value += ctx.source.substring(pos, nl + 1);
                        pos = i + 1;
                    }
                }
                break;
        }
    }
    value += ctx.source.substring(pos, i);
    return { type: 'text', start, end: i, value };
}
exports.parseText = parseText;
function parseLiteral(ctx, start, required) {
    if (ctx.source[start] === '|')
        return parseQuotedLiteral(ctx, start);
    const value = (0, names_js_1.parseUnquotedLiteralValue)(ctx.source, start);
    if (!value) {
        if (required)
            ctx.onError('empty-token', start, start);
        else
            return undefined;
    }
    const end = start + value.length;
    return { type: 'literal', quoted: false, start, end, value };
}
exports.parseLiteral = parseLiteral;
function parseQuotedLiteral(ctx, start) {
    let value = '';
    let pos = start + 1;
    const open = { start, end: pos, value: '|' };
    for (let i = pos; i < ctx.source.length; ++i) {
        switch (ctx.source[i]) {
            case '\\': {
                const esc = parseEscape(ctx, 'literal', i);
                if (esc) {
                    value += ctx.source.substring(pos, i) + esc.value;
                    i += esc.length;
                    pos = i + 1;
                }
                break;
            }
            case '|':
                value += ctx.source.substring(pos, i);
                return {
                    type: 'literal',
                    quoted: true,
                    start,
                    end: i + 1,
                    open,
                    value,
                    close: { start: i, end: i + 1, value: '|' }
                };
            case '\n':
                if (ctx.resource) {
                    const nl = i;
                    let next = ctx.source[i + 1];
                    while (next === ' ' || next === '\t') {
                        i += 1;
                        next = ctx.source[i + 1];
                    }
                    if (i > nl) {
                        value += ctx.source.substring(pos, nl + 1);
                        pos = i + 1;
                    }
                }
                break;
        }
    }
    value += ctx.source.substring(pos);
    ctx.onError('missing-syntax', ctx.source.length, '|');
    return {
        type: 'literal',
        quoted: true,
        start,
        end: ctx.source.length,
        open,
        value,
        close: undefined
    };
}
exports.parseQuotedLiteral = parseQuotedLiteral;
function parseVariable(ctx, start) {
    const pos = start + 1;
    const open = { start, end: pos, value: '$' };
    const name = (0, names_js_1.parseNameValue)(ctx.source, pos);
    const end = pos + name.length;
    if (!name)
        ctx.onError('empty-token', pos, pos + 1);
    return { type: 'variable', start, end, open, name };
}
exports.parseVariable = parseVariable;
function parseEscape(ctx, scope, start) {
    const raw = ctx.source[start + 1];
    switch (raw) {
        case '\\':
            return { value: raw, length: 1 };
        case '{':
        case '}':
            if (scope === 'text')
                return { value: raw, length: 1 };
            break;
        case '|':
            if (scope === 'literal')
                return { value: raw, length: 1 };
            break;
        default:
            if (ctx.resource) {
                let hexLen = 0;
                switch (raw) {
                    case '\t':
                    case ' ':
                        return { value: raw, length: 1 };
                    case 'n':
                        return { value: '\n', length: 1 };
                    case 'r':
                        return { value: '\r', length: 1 };
                    case 't':
                        return { value: '\t', length: 1 };
                    case 'u':
                        hexLen = 4;
                        break;
                    case 'U':
                        hexLen = 6;
                        break;
                    case 'x':
                        hexLen = 2;
                        break;
                }
                if (hexLen > 0) {
                    const h0 = start + 2;
                    const raw = ctx.source.substring(h0, h0 + hexLen);
                    if (raw.length === hexLen && /^[0-9A-Fa-f]+$/.test(raw)) {
                        return {
                            value: String.fromCharCode(parseInt(raw, 16)),
                            length: 1 + hexLen
                        };
                    }
                }
            }
    }
    ctx.onError('bad-escape', start, start + 2);
    return null;
}
