"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.messageFromCST = exports.cst = void 0;
const errors_js_1 = require("../errors.js");
/**
 * Shared symbol used as a key on message data model nodes
 * to reference their CST source.
 *
 * @beta
 */
exports.cst = Symbol.for('CST');
/**
 * Convert a CST message structure into its data model representation.
 *
 * @beta
 */
function messageFromCST(msg) {
    for (const error of msg.errors)
        throw error;
    const declarations = msg.declarations
        ? msg.declarations.map(asDeclaration)
        : [];
    if (msg.type === 'select') {
        return {
            type: 'select',
            declarations,
            selectors: msg.selectors.map(sel => asExpression(sel, false)),
            variants: msg.variants.map(variant => ({
                keys: variant.keys.map(key => key.type === '*' ? { type: '*', [exports.cst]: key } : asValue(key)),
                value: asPattern(variant.value),
                [exports.cst]: variant
            })),
            [exports.cst]: msg
        };
    }
    else {
        return {
            type: 'message',
            declarations,
            pattern: asPattern(msg.pattern),
            [exports.cst]: msg
        };
    }
}
exports.messageFromCST = messageFromCST;
function asDeclaration(decl) {
    switch (decl.type) {
        case 'input': {
            const value = asExpression(decl.value, false);
            if (value.arg?.type !== 'variable') {
                const { start, end } = decl.value;
                throw new errors_js_1.MessageSyntaxError('parse-error', start, end);
            }
            return {
                type: 'input',
                name: value.arg.name,
                value: value,
                [exports.cst]: decl
            };
        }
        case 'local':
            return {
                type: 'local',
                name: asValue(decl.target).name,
                value: asExpression(decl.value, false),
                [exports.cst]: decl
            };
        default:
            return {
                type: 'unsupported-statement',
                keyword: (decl.keyword?.value ?? '').substring(1),
                body: decl.body?.value || undefined,
                expressions: decl.values?.map(dv => asExpression(dv, true)) ?? [],
                [exports.cst]: decl
            };
    }
}
const asPattern = (pattern) => pattern.body.map(el => el.type === 'text' ? el.value : asExpression(el, true));
function asExpression(exp, allowMarkup) {
    if (exp.type === 'expression') {
        const attributes = exp.attributes.length
            ? exp.attributes.map(asAttribute)
            : undefined;
        if (allowMarkup && exp.markup) {
            const cm = exp.markup;
            const name = asName(cm.name);
            const kind = cm.open.value === '/' ? 'close' : cm.close ? 'standalone' : 'open';
            const markup = { type: 'markup', kind, name };
            if (cm.options.length)
                markup.options = cm.options.map(asOption);
            if (attributes)
                markup.attributes = attributes;
            markup[exports.cst] = exp;
            return markup;
        }
        const arg = exp.arg ? asValue(exp.arg) : undefined;
        let annotation;
        const ca = exp.annotation;
        if (ca) {
            switch (ca.type) {
                case 'function':
                    annotation = { type: 'function', name: asName(ca.name) };
                    if (ca.options.length)
                        annotation.options = ca.options.map(asOption);
                    break;
                case 'reserved-annotation':
                    annotation = {
                        type: 'unsupported-annotation',
                        source: ca.open.value + ca.source.value
                    };
                    break;
                default:
                    throw new errors_js_1.MessageSyntaxError('parse-error', exp.start, exp.end);
            }
        }
        let expression = arg
            ? { type: 'expression', arg }
            : undefined;
        if (annotation) {
            annotation[exports.cst] = ca;
            if (expression)
                expression.annotation = annotation;
            else
                expression = { type: 'expression', annotation };
        }
        if (expression) {
            if (attributes)
                expression.attributes = attributes;
            expression[exports.cst] = exp;
            return expression;
        }
    }
    throw new errors_js_1.MessageSyntaxError('parse-error', exp.start, exp.end);
}
const asOption = (option) => ({
    name: asName(option.name),
    value: asValue(option.value),
    [exports.cst]: option
});
function asAttribute(attr) {
    const name = asName(attr.name);
    return attr.value
        ? { name, value: asValue(attr.value), [exports.cst]: attr }
        : { name, [exports.cst]: attr };
}
function asName(id) {
    switch (id.length) {
        case 1:
            return id[0].value;
        case 3:
            return `${id[0].value}:${id[2].value}`;
        default:
            throw new errors_js_1.MessageSyntaxError('parse-error', id[0]?.start ?? -1, id.at(-1)?.end ?? -1);
    }
}
function asValue(value) {
    switch (value.type) {
        case 'literal':
            return { type: 'literal', value: value.value, [exports.cst]: value };
        case 'variable':
            return { type: 'variable', name: value.name, [exports.cst]: value };
        default:
            throw new errors_js_1.MessageSyntaxError('parse-error', value.start, value.end);
    }
}
