"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseMessage = void 0;
const names_js_1 = require("../cst/names.js");
const errors_js_1 = require("../errors.js");
const whitespaceChars = ['\t', '\n', '\r', ' ', '\u3000'];
//// Parser State ////
let pos;
let source;
let opt;
//// Utilities & Error Wrappers ////
// These indirections allow for the function names to be mangled,
// while keeping the error class name intact.
const MissingSyntax = (pos, expected) => new errors_js_1.MessageSyntaxError('missing-syntax', pos, pos + expected.length, expected);
const SyntaxError = (...args) => new errors_js_1.MessageSyntaxError(...args);
function expect(searchString, consume) {
    if (source.startsWith(searchString, pos)) {
        if (consume)
            pos += searchString.length;
    }
    else {
        throw MissingSyntax(pos, searchString);
    }
}
function parseMessage(source_, opt_ = {}) {
    pos = 0;
    source = source_;
    opt = opt_;
    const decl = declarations();
    if (source.startsWith('.match', pos))
        return selectMessage(decl);
    const quoted = decl.length > 0 || source.startsWith('{{', pos);
    const pattern_ = pattern(quoted);
    if (quoted) {
        ws();
        if (pos < source.length) {
            throw SyntaxError('extra-content', pos, source.length);
        }
    }
    return { type: 'message', declarations: decl, pattern: pattern_ };
}
exports.parseMessage = parseMessage;
function selectMessage(declarations) {
    pos += 6; // '.match'
    ws();
    const selectors = [];
    while (source[pos] === '{') {
        selectors.push(expression(false));
        ws();
    }
    if (selectors.length === 0)
        throw SyntaxError('empty-token', pos);
    const variants = [];
    while (pos < source.length) {
        variants.push(variant());
        ws();
    }
    return { type: 'select', declarations, selectors, variants };
}
function variant() {
    const keys = [];
    while (pos < source.length) {
        ws(keys.length ? '{' : false);
        const next = source[pos];
        if (next === '{')
            break;
        if (next === '*') {
            keys.push({ type: '*' });
            pos += 1;
        }
        else {
            keys.push(literal(true));
        }
    }
    return { keys, value: pattern(true) };
}
function pattern(quoted) {
    if (quoted) {
        if (source.startsWith('{{', pos))
            pos += 2;
        else
            throw MissingSyntax(pos, '{{');
    }
    const pattern = [];
    loop: while (pos < source.length) {
        switch (source[pos]) {
            case '{': {
                pattern.push(expression(true));
                break;
            }
            case '}':
                if (!quoted)
                    throw SyntaxError('parse-error', pos);
                break loop;
            default: {
                pattern.push(text());
            }
        }
    }
    if (quoted) {
        if (source.startsWith('}}', pos))
            pos += 2;
        else
            throw MissingSyntax(pos, '}}');
    }
    return pattern;
}
function declarations() {
    const declarations = [];
    loop: while (source[pos] === '.') {
        const keyword = (0, names_js_1.parseNameValue)(source, pos + 1);
        switch (keyword) {
            case 'input':
                declarations.push(inputDeclaration());
                break;
            case 'local':
                declarations.push(localDeclaration());
                break;
            case 'match':
                break loop;
            case '':
                throw SyntaxError('parse-error', pos);
            default:
                declarations.push(unsupportedStatement(keyword));
        }
        ws();
    }
    return declarations;
}
function inputDeclaration() {
    pos += 6; // '.input'
    ws();
    expect('{', false);
    const valueStart = pos;
    const value = expression(false);
    if (value.type === 'expression' && value.arg?.type === 'variable') {
        // @ts-expect-error TS isn't catching that value is Expression<VariableRef>
        return { type: 'input', name: value.arg.name, value };
    }
    throw SyntaxError('bad-input-expression', valueStart, pos);
}
function localDeclaration() {
    pos += 6; // '.local'
    ws(true);
    expect('$', true);
    const name_ = name();
    ws();
    expect('=', true);
    ws();
    expect('{', false);
    const value = expression(false);
    return { type: 'local', name: name_, value };
}
function unsupportedStatement(keyword) {
    pos += 1 + keyword.length; // '.' + keyword
    ws('{');
    const body = reservedBody();
    const expressions = [];
    while (source[pos] === '{') {
        if (source.startsWith('{{', pos))
            break;
        expressions.push(expression(false));
        ws();
    }
    if (expressions.length === 0)
        throw SyntaxError('empty-token', pos);
    return { type: 'unsupported-statement', keyword, body, expressions };
}
function expression(allowMarkup) {
    const start = pos;
    pos += 1; // '{'
    ws();
    const arg = value(false);
    if (arg)
        ws('}');
    const sigil = source[pos];
    let annotation;
    let markup;
    switch (sigil) {
        case '@':
        case '}':
            break;
        case ':': {
            pos += 1; // ':'
            annotation = { type: 'function', name: identifier() };
            const options_ = options();
            if (options_.length)
                annotation.options = options_;
            break;
        }
        case '#':
        case '/': {
            if (arg || !allowMarkup)
                throw SyntaxError('parse-error', pos);
            pos += 1; // '#' or '/'
            const kind = sigil === '#' ? 'open' : 'close';
            markup = { type: 'markup', kind, name: identifier() };
            const options_ = options();
            if (options_.length)
                markup.options = options_;
            break;
        }
        case '^':
        case '&':
            annotation = privateAnnotation(sigil);
            break;
        case '!':
        case '%':
        case '*':
        case '+':
        case '<':
        case '>':
        case '?':
        case '~':
            annotation = unsupportedAnnotation(sigil);
            break;
        default:
            throw SyntaxError('parse-error', pos);
    }
    while (source[pos] === '@')
        attribute();
    if (markup?.kind === 'open' && source[pos] === '/') {
        markup.kind = 'standalone';
        pos += 1; // '/'
    }
    expect('}', true);
    if (annotation) {
        return arg
            ? { type: 'expression', arg, annotation }
            : { type: 'expression', annotation };
    }
    if (markup)
        return markup;
    if (!arg)
        throw SyntaxError('empty-token', start, pos);
    return { type: 'expression', arg };
}
/** Requires and consumes leading and trailing whitespace. */
function options() {
    ws('/}');
    const options = [];
    while (pos < source.length) {
        const next = source[pos];
        if (next === '@' || next === '/' || next === '}')
            break;
        const name_ = identifier();
        ws();
        expect('=', true);
        ws();
        options.push({ name: name_, value: value(true) });
        ws('/}');
    }
    return options;
}
function attribute() {
    pos += 1; // '@'
    identifier(); // name
    ws('=/}');
    if (source[pos] === '=') {
        pos += 1; // '='
        ws();
        value(true); // value
        ws('/}');
    }
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function privateAnnotation(sigil) {
    if (opt.privateAnnotation) {
        const res = opt.privateAnnotation(source, pos);
        pos = res.pos;
        ws('}');
        return res.annotation;
    }
    return unsupportedAnnotation(sigil);
}
function unsupportedAnnotation(sigil) {
    pos += 1; // sigil
    return { type: 'unsupported-annotation', source: sigil + reservedBody() };
}
function reservedBody() {
    const start = pos;
    loop: while (pos < source.length) {
        const next = source[pos];
        switch (next) {
            case '\\': {
                switch (source[pos + 1]) {
                    case '\\':
                    case '{':
                    case '|':
                    case '}':
                        break;
                    default:
                        throw SyntaxError('bad-escape', pos, pos + 2);
                }
                pos += 2;
                break;
            }
            case '|':
                quotedLiteral();
                break;
            case '@':
                pos -= 1;
                ws(true);
                break loop;
            case '{':
            case '}':
                break loop;
            default: {
                const cc = next.charCodeAt(0);
                if (cc >= 0xd800 && cc < 0xe000) {
                    // surrogates are invalid here
                    throw SyntaxError('parse-error', pos);
                }
                pos += 1;
            }
        }
    }
    return source.substring(start, pos).trimEnd();
}
function text() {
    let value = '';
    let i = pos;
    loop: for (; i < source.length; ++i) {
        switch (source[i]) {
            case '\\': {
                const esc = source[i + 1];
                if (esc !== '\\' && esc !== '{' && esc !== '}') {
                    throw SyntaxError('bad-escape', i, i + 2);
                }
                value += source.substring(pos, i) + esc;
                i += 1;
                pos = i + 1;
                break;
            }
            case '{':
            case '}':
                break loop;
        }
    }
    value += source.substring(pos, i);
    pos = i;
    return value;
}
function value(required) {
    if (source[pos] === '$') {
        pos += 1; // '$'
        return { type: 'variable', name: name() };
    }
    else {
        return literal(required);
    }
}
function literal(required) {
    if (source[pos] === '|')
        return quotedLiteral();
    const value = (0, names_js_1.parseUnquotedLiteralValue)(source, pos);
    if (!value) {
        if (required)
            throw SyntaxError('empty-token', pos);
        else
            return undefined;
    }
    pos += value.length;
    return { type: 'literal', value };
}
function quotedLiteral() {
    pos += 1; // '|'
    let value = '';
    for (let i = pos; i < source.length; ++i) {
        switch (source[i]) {
            case '\\': {
                const esc = source[i + 1];
                if (esc !== '\\' && esc !== '|') {
                    throw SyntaxError('bad-escape', i, i + 2);
                }
                value += source.substring(pos, i) + esc;
                i += 1;
                pos = i + 1;
                break;
            }
            case '|':
                value += source.substring(pos, i);
                pos = i + 1;
                return { type: 'literal', value };
        }
    }
    throw MissingSyntax(source.length, '|');
}
function identifier() {
    const name_ = name();
    if (source[pos] === ':') {
        pos += 1;
        return name_ + ':' + name();
    }
    return name_;
}
function name() {
    const name = (0, names_js_1.parseNameValue)(source, pos);
    if (!name)
        throw SyntaxError('empty-token', pos);
    pos += name.length;
    return name;
}
function ws(req = false) {
    let length = 0;
    let next = source[pos];
    while (whitespaceChars.includes(next)) {
        length += 1;
        next = source[pos + length];
    }
    pos += length;
    if (req && !length && (req === true || !req.includes(source[pos]))) {
        throw MissingSyntax(pos, "' '");
    }
}
