"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveVariableRef = exports.lookupVariableRef = exports.UnresolvedExpression = void 0;
const errors_js_1 = require("../errors.js");
const fallback_js_1 = require("../functions/fallback.js");
const unknown_js_1 = require("../functions/unknown.js");
const function_context_js_1 = require("./function-context.js");
const resolve_expression_js_1 = require("./resolve-expression.js");
/**
 * Declarations aren't resolved until they're requierd,
 * and their resolution order matters for variable resolution.
 * This internal class is used to store any required data,
 * and to allow for `instanceof` detection.
 * @private
 */
class UnresolvedExpression {
    expression;
    scope;
    constructor(expression, scope) {
        this.expression = expression;
        this.scope = scope;
    }
}
exports.UnresolvedExpression = UnresolvedExpression;
const isScope = (scope) => scope instanceof Object;
/**
 * Looks for the longest matching `.` delimited starting substring of name.
 * @returns `undefined` if value not found
 */
function getValue(scope, name) {
    if (isScope(scope)) {
        if (name in scope)
            return scope[name];
        const parts = name.split('.');
        for (let i = parts.length - 1; i > 0; --i) {
            const head = parts.slice(0, i).join('.');
            if (head in scope) {
                const tail = parts.slice(i).join('.');
                return getValue(scope[head], tail);
            }
        }
    }
    return undefined;
}
/**
 * Get the raw value of a variable.
 * Resolves declarations as necessary
 *
 * @internal
 * @returns `unknown` or `any` for input values;
 *   `MessageValue` for `.input` and `.local` declaration values.
 */
function lookupVariableRef(ctx, { name }) {
    const value = getValue(ctx.scope, name);
    if (value === undefined) {
        const source = '$' + name;
        const msg = `Variable not available: ${source}`;
        ctx.onError(new errors_js_1.MessageResolutionError('unresolved-var', msg, source));
    }
    else if (value instanceof UnresolvedExpression) {
        const local = (0, resolve_expression_js_1.resolveExpression)(value.scope ? { ...ctx, scope: value.scope } : ctx, value.expression);
        ctx.scope[name] = local;
        ctx.localVars.add(local);
        return local;
    }
    return value;
}
exports.lookupVariableRef = lookupVariableRef;
function resolveVariableRef(ctx, ref) {
    const source = '$' + ref.name;
    const value = lookupVariableRef(ctx, ref);
    let type = typeof value;
    if (type === 'object') {
        const mv = value;
        if (ctx.localVars.has(mv))
            return mv;
        if (value instanceof Number)
            type = 'number';
        else if (value instanceof String)
            type = 'string';
    }
    switch (type) {
        case 'bigint':
        case 'number': {
            const msgCtx = new function_context_js_1.MessageFunctionContext(ctx, source);
            return ctx.functions.number(msgCtx, {}, value);
        }
        case 'string': {
            const msgCtx = new function_context_js_1.MessageFunctionContext(ctx, source);
            return ctx.functions.string(msgCtx, {}, value);
        }
    }
    return value === undefined ? (0, fallback_js_1.fallback)(source) : (0, unknown_js_1.unknown)(source, value);
}
exports.resolveVariableRef = resolveVariableRef;
