"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.stringifyMessage = void 0;
const messageformat_js_1 = require("../messageformat.js");
const names_js_1 = require("../cst/names.js");
const type_guards_js_1 = require("./type-guards.js");
/**
 * Stringify a message using its syntax representation.
 *
 * @beta
 */
function stringifyMessage(msg) {
    if (msg instanceof messageformat_js_1.MessageFormat)
        msg = msg.resolvedOptions().message;
    let res = '';
    for (const decl of msg.declarations)
        res += stringifyDeclaration(decl);
    if ((0, type_guards_js_1.isPatternMessage)(msg)) {
        res += stringifyPattern(msg.pattern, !!res);
    }
    else if ((0, type_guards_js_1.isSelectMessage)(msg)) {
        res += '.match';
        for (const sel of msg.selectors)
            res += ' ' + stringifyExpression(sel);
        for (const { keys, value } of msg.variants) {
            res += '\n';
            for (const key of keys) {
                res += ((0, type_guards_js_1.isLiteral)(key) ? stringifyLiteral(key) : '*') + ' ';
            }
            res += stringifyPattern(value, true);
        }
    }
    return res;
}
exports.stringifyMessage = stringifyMessage;
function stringifyDeclaration(decl) {
    switch (decl.type) {
        case 'input':
            return `.input ${stringifyExpression(decl.value)}\n`;
        case 'local':
            return `.local $${decl.name} = ${stringifyExpression(decl.value)}\n`;
        case 'unsupported-statement': {
            const parts = [`.${decl.keyword}`];
            if (decl.body)
                parts.push(decl.body);
            for (const exp of decl.expressions) {
                parts.push(exp.type === 'expression'
                    ? stringifyExpression(exp)
                    : stringifyMarkup(exp));
            }
            return parts.join(' ');
        }
    }
    // @ts-expect-error Guard against non-TS users with bad data
    throw new Error(`Unsupported ${decl.type} declaration`);
}
function stringifyFunctionAnnotation({ name, options }) {
    let res = `:${name}`;
    if (options)
        for (const opt of options)
            res += ' ' + stringifyOption(opt);
    return res;
}
function stringifyMarkup({ kind, name, options, attributes }) {
    let res = kind === 'close' ? '{/' : '{#';
    res += name;
    if (options)
        for (const opt of options)
            res += ' ' + stringifyOption(opt);
    if (attributes) {
        for (const attr of attributes)
            res += ' ' + stringifyAttribute(attr);
    }
    res += kind === 'standalone' ? ' /}' : '}';
    return res;
}
function stringifyLiteral({ value }) {
    if ((0, names_js_1.isValidUnquotedLiteral)(value))
        return value;
    const esc = value.replace(/\\/g, '\\\\').replace(/\|/g, '\\|');
    return `|${esc}|`;
}
function stringifyOption({ name, value }) {
    const valueStr = (0, type_guards_js_1.isVariableRef)(value)
        ? stringifyVariableRef(value)
        : stringifyLiteral(value);
    return `${name}=${valueStr}`;
}
function stringifyAttribute({ name, value }) {
    if (!value)
        return `@${name}`;
    const valueStr = (0, type_guards_js_1.isVariableRef)(value)
        ? stringifyVariableRef(value)
        : stringifyLiteral(value);
    return `@${name}=${valueStr}`;
}
function stringifyPattern(pattern, quoted) {
    let res = '';
    if (!quoted && typeof pattern[0] === 'string' && pattern[0][0] === '.') {
        quoted = true;
    }
    for (const el of pattern) {
        if (typeof el === 'string')
            res += el;
        else if (el.type === 'markup')
            res += stringifyMarkup(el);
        else
            res += stringifyExpression(el);
    }
    return quoted ? `{{${res}}}` : res;
}
function stringifyExpression({ arg, annotation, attributes }) {
    let res;
    switch (arg?.type) {
        case 'literal':
            res = stringifyLiteral(arg);
            break;
        case 'variable':
            res = stringifyVariableRef(arg);
            break;
        default:
            res = '';
    }
    if (annotation) {
        if (res)
            res += ' ';
        res +=
            annotation.type === 'function'
                ? stringifyFunctionAnnotation(annotation)
                : annotation.source ?? '�';
    }
    if (attributes) {
        for (const attr of attributes)
            res += ' ' + stringifyAttribute(attr);
    }
    return `{${res}}`;
}
function stringifyVariableRef(ref) {
    return '$' + ref.name;
}
