"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.visit = void 0;
/**
 * Apply visitor functions to message nodes.
 *
 * @remarks
 * The visitors are applied in source order, starting from the root.
 * Visitors for nodes that contain other nodes may return a callback function
 * that will be called with no arguments when exiting the node.
 *
 * If set, the `node` visitor is called for all {@link MessageNode} values
 * for which an explicit visitor is not defined.
 *
 * Many visitors will be called with additional arguments
 * identifying some of the context for the visited node.
 *
 * @beta
 */
function visit(msg, visitors) {
    const { node, pattern } = visitors;
    const { annotation = node, attribute = node, declaration = node, expression = node, key = node, markup = node, option = node, value = node, variant = node } = visitors;
    const handleOptions = (options, context) => {
        if (options) {
            for (let i = 0; i < options.length; ++i) {
                const opt = options[i];
                const end = option?.(opt, i, options, context);
                value?.(opt.value, context, 'option');
                end?.();
            }
        }
    };
    const handleAttributes = (attributes, context) => {
        if (attributes) {
            for (let i = 0; i < attributes.length; ++i) {
                const attr = attributes[i];
                const end = attribute?.(attr, i, attributes, context);
                if (attr.value)
                    value?.(attr.value, context, 'attribute');
                end?.();
            }
        }
    };
    const handleElement = (exp, context) => {
        if (typeof exp === 'object') {
            let end;
            switch (exp.type) {
                case 'expression': {
                    end = expression?.(exp, context);
                    if (exp.arg)
                        value?.(exp.arg, context, 'arg');
                    if (exp.annotation) {
                        const endA = annotation?.(exp.annotation, context, exp.arg);
                        handleOptions(exp.annotation.options, context);
                        endA?.();
                    }
                    handleAttributes(exp.attributes, context);
                    break;
                }
                case 'markup': {
                    end = context !== 'selector' ? markup?.(exp, context) : undefined;
                    handleOptions(exp.options, context);
                    handleAttributes(exp.attributes, context);
                    break;
                }
            }
            end?.();
        }
    };
    const handlePattern = (pat) => {
        const end = pattern?.(pat);
        for (const el of pat)
            handleElement(el, 'placeholder');
        end?.();
    };
    for (const decl of msg.declarations) {
        const end = declaration?.(decl);
        if (decl.value)
            handleElement(decl.value, 'declaration');
        else
            for (const exp of decl.expressions)
                handleElement(exp, 'declaration');
        end?.();
    }
    if (msg.type === 'message') {
        handlePattern(msg.pattern);
    }
    else {
        for (const sel of msg.selectors)
            handleElement(sel, 'selector');
        for (const vari of msg.variants) {
            const end = variant?.(vari);
            if (key)
                vari.keys.forEach(key);
            handlePattern(vari.value);
            end?.();
        }
    }
}
exports.visit = visit;
