"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.time = exports.date = exports.datetime = void 0;
const errors_js_1 = require("../errors.js");
const utils_js_1 = require("./utils.js");
const localeOptions = [
    'calendar',
    'localeMatcher',
    'hour12',
    'hourCycle',
    'numberingSystem',
    'timeZone'
];
/**
 * `datetime` accepts a Date, number or string as its input
 * and formats it with the same options as
 * {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/DateTimeFormat | Intl.DateTimeFormat}.
 *
 * @beta
 */
const datetime = (ctx, options, input) => dateTimeImplementation(ctx, options, input, res => {
    for (const [name, value] of Object.entries(options)) {
        if (value === undefined)
            continue;
        try {
            switch (name) {
                case 'locale':
                    break;
                case 'fractionalSecondDigits':
                    res[name] = (0, utils_js_1.asPositiveInteger)(value);
                    break;
                case 'hour12':
                    res[name] = (0, utils_js_1.asBoolean)(value);
                    break;
                default:
                    res[name] = (0, utils_js_1.asString)(value);
            }
        }
        catch {
            const msg = `Value ${value} is not valid for :datetime option ${name}`;
            throw new errors_js_1.MessageResolutionError('bad-option', msg, ctx.source);
        }
    }
    // Set defaults if localeMatcher is the only option
    if (Object.keys(res).length <= 1) {
        res.dateStyle = 'short';
        res.timeStyle = 'short';
    }
});
exports.datetime = datetime;
/**
 * `date` accepts a Date, number or string as its input
 * and formats it according to a single "style" option.
 *
 * @beta
 */
const date = (ctx, options, input) => dateTimeImplementation(ctx, options, input, res => {
    const ds = options.style ?? res.dateStyle ?? 'short';
    for (const name of Object.keys(res)) {
        if (!localeOptions.includes(name))
            delete res[name];
    }
    res.dateStyle = (0, utils_js_1.asString)(ds);
});
exports.date = date;
/**
 * `time` accepts a Date, number or string as its input
 * and formats it according to a single "style" option.
 *
 * @beta
 */
const time = (ctx, options, input) => dateTimeImplementation(ctx, options, input, res => {
    const ts = options.style ?? res.timeStyle ?? 'short';
    for (const name of Object.keys(res)) {
        if (!localeOptions.includes(name))
            delete res[name];
    }
    res.timeStyle = (0, utils_js_1.asString)(ts);
});
exports.time = time;
function dateTimeImplementation({ localeMatcher, locales, source }, options, input, parseOptions) {
    const lc = (0, utils_js_1.mergeLocales)(locales, input, options);
    const opt = { localeMatcher };
    if (input && typeof input === 'object') {
        if (input && 'options' in input)
            Object.assign(opt, input.options);
        if (!(input instanceof Date) && typeof input.valueOf === 'function') {
            input = input.valueOf();
        }
    }
    let value;
    switch (typeof input) {
        case 'number':
        case 'string':
            value = new Date(input);
            break;
        case 'object':
            value = input;
            break;
    }
    if (!(value instanceof Date)) {
        const msg = 'Input is not a date';
        throw new errors_js_1.MessageResolutionError('bad-input', msg, source);
    }
    parseOptions(opt);
    const date = value;
    let locale;
    let dtf;
    let str;
    return {
        type: 'datetime',
        source,
        get locale() {
            return (locale ??= Intl.DateTimeFormat.supportedLocalesOf(lc, opt)[0]);
        },
        get options() {
            return { ...opt };
        },
        toParts() {
            dtf ??= new Intl.DateTimeFormat(lc, opt);
            const parts = dtf.formatToParts(date);
            locale ??= dtf.resolvedOptions().locale;
            return [{ type: 'datetime', source, locale, parts }];
        },
        toString() {
            dtf ??= new Intl.DateTimeFormat(lc, opt);
            str ??= dtf.format(date);
            return str;
        },
        valueOf: () => date
    };
}
